def extract_anime_details(soup, mal_id):
    """
    Extracts an anime page's details and statistics.

    Parameters
    ----------
    soup: BeautifulSoup object
        The BeautifulSoup object containing the page's content
    mal_id: int
        Id of the anime being scraped
    """
    dark_text_child_data = soup.find_all("span", class_="dark_text")
    details_dict = {}

    for child in dark_text_child_data:
        parent = child.parent

        if parent.has_attr("itemprop"):
            rating_value = parent.find(
                "span", itemprop="ratingValue").text.strip()
            rating_count = parent.find(
                "span", itemprop="ratingCount").text.strip()
            details_dict["Score"] = rating_value
            details_dict["ScoredByCount"] = rating_count
        else:
            superscript_element = parent.find("sup")
            statistics_element = parent.find("div", class_="statistics-info info2")

            # Removing uninteresting elements to make data easier to parse
            if superscript_element is not None:
                superscript_element.decompose()

            if statistics_element is not None:
                statistics_element.decompose()

            key_value = parent.text.strip().split(":")
            details_dict[key_value[0]] = key_value[1]

    title = soup.select("h1.title-name strong")[0].text.strip()
    description = soup.find("p", itemprop="description").text.strip()
    image = soup.find("img", itemprop="image")["data-src"].strip()
    video = soup.find("a", class_="iframe js-fancybox-video video-unit promotion")
    
    if video is not None:
        promo_video = video["href"]
        promo_video_background_image = video["style"]
        details_dict["PromoVideo"] = promo_video
        details_dict["PromoVideoBackgroundImage"] = promo_video_background_image

    details_dict["Title"] = title
    details_dict["ImageSrc"] = image
    details_dict["Synopsis"] = description
    details_dict["MyAnimeListId"] = mal_id

    return details_dict


def create_mal_anime_details_table():
    """
    Returns query string for creating the MAL_Anime_Details table.
    """

    return """
    CREATE TABLE MAL_Anime_Details (
        MAL_Id int NOT NULL PRIMARY KEY,
        Title varchar(255) DEFAULT NULL,
        AltNameEnglish varchar(255) DEFAULT NULL,
        AltNameSynonyms varchar(255) DEFAULT NULL,
        AltNameJapanese varchar(255) DEFAULT NULL,
        Synopsis mediumtext DEFAULT NULL,
        MediaType varchar(50) DEFAULT NULL,
        EpisodeCount int DEFAULT NULL,
        CurrentStatus varchar(255) DEFAULT NULL,
        Aired varchar(255) DEFAULT NULL,
        Premiered varchar(255) DEFAULT NULL,
        Broadcast varchar(255) DEFAULT NULL,
        Producers varchar(255) DEFAULT NULL,
        Licensors varchar(255) DEFAULT NULL,
        Studios varchar(255) DEFAULT NULL,
        Source varchar(50) DEFAULT NULL,
        Genres varchar(255) DEFAULT NULL,
        Duration varchar(255) DEFAULT NULL,
        Rating varchar(50) DEFAULT NULL,
        PromoVideoSrc varchar(255) DEFAULT NULL,
        PromoVideoBackgroundImageSrc varchar(255) DEFAULT NULL,
        ImageSrc varchar(255) DEFAULT NULL
    );
    """

def create_mal_anime_detail_stats_table():
    """
    Returns query string for creating the MAL_Anime_Detail_Stats table.
    """

    return """
    CREATE TABLE MAL_Anime_Detail_Stats (
        MAL_Id int NOT NULL PRIMARY KEY,
        Score float(24) DEFAULT NULL,
        ScoredByCount int DEFAULT NULL,
        Ranked int DEFAULT NULL,
        Popularity int DEFAULT NULL,
        MembersCount int DEFAULT NULL,
        FavoritesCount int DEFAULT NULL,
        FOREIGN KEY (MAL_Id) REFERENCES MAL_Anime_Details(MAL_Id)
    );
    """

def insert_mal_anime_detail(data):
    """
    Returns tuple (query, values) necessary for inseting 
    data into the MAL_Anime_Details table.
    """

    query = """
    INSERT INTO MAL_Anime_Details 
    (
        MAL_Id, 
        Title, 
        AltNameEnglish,
        AltNameSynonyms,
        AltNameJapanese,
        Synopsis,
        MediaType,
        EpisodeCount,
        CurrentStatus,
        Aired,
        Premiered,
        Broadcast,
        Producers,
        Licensors,
        Studios,
        Source,
        Genres,
        Duration,
        Rating,
        PromoVideoSrc,
        PromoVideoBackgroundImageSrc,
        ImageSrc
    )
    VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s,
        %s, %s,%s, %s, %s, %s, %s, %s, %s, %s, %s, %s)
    """

    values = (
        data["MyAnimeListId"],
        data["Title"],
        data["AltNameEnglish"],
        data["AltNameSynonyms"],
        data["AltNameJapanese"],
        data["Synopsis"],
        data["MediaType"],
        data["EpisodeCount"],
        data["CurrentStatus"],
        data["Aired"],
        data["Premiered"],
        data["Broadcast"],
        data["Producers"],
        data["Licensors"],
        data["Studios"],
        data["Source"],
        data["Genres"],
        data["Duration"],
        data["Rating"],
        data["PromoVideo"],
        data["PromoVideoBackgroundImage"],
        data["ImageSrc"]
    )

    return query, values

def insert_mal_anime_detail_stats(data):
    """
    Returns tuple (query, values) necessary for inseting 
    data into the MAL_Anime_Detail_Stats table.
    """

    query = """
    INSERT INTO MAL_Anime_Detail_Stats
    (
        MAL_Id,
        Score,
        ScoredByCount,
        Ranked,
        Popularity,
        MembersCount,
        FavoritesCount
    )
    VALUES (%s, %s, %s, %s, %s, %s, %s)
    """

    values = (
        data["MyAnimeListId"],
        data["Score"],
        data["ScoredByCount"],
        data["Ranked"],
        data["Popularity"],
        data["MembersCount"],
        data["FavoritesCount"]
    )

    return query, values


import mysql.connector

class DatabaseManager:
    """
    This class defines a database manager for opening connections, 
    closing connections, and querying the tables within the database.
    """

    def __init__(self, host, user, password, database):
        """
        Parameters
        ----------
        host: str
            The database host.
        user: str
            The database user.
        password: str
            The database password.
        database: str
            The database name.
        """
        self.host = host
        self.user = user
        self.password = password
        self.database = database
        self.db = mysql.connector.connect(
            host=host,
            user=user,
            password=password,
            database=database
        )

    def close(self):
        """
        Closes the open connection to the database
        Users of this class should **always** call close when they are
        done with this class so it can clean up the DB connection.
        """
        self.db.close()

    def connect(self):
        """
        Opens a connection to the database.
        """
        self.db = mysql.connector.connect(
            host=self.host,
            user=self.user,
            password=self.password,
            database=self.database
        )

    def execute(self, query, values):
        """
        Executes the provided query with the value(s) on the 
        database.

        Parameters
        ----------
        query: str
            A sql insert query.
        values: tuple
            Values to be inserted to the table
        """
        cursor = self.db.cursor()

        if type(values) is list:
            cursor.executemany(query, values)
        elif type(values) is None:
            cursor.execute(query)
        else:
            cursor.execute(query, values)

        self.db.commit()

    def check_table_exists(self, tablename):
        """
        Checks if a table exists within the database.

        tablename: str
            Name of the table to check if exists.
        """

        cursor = self.db.cursor()

        cursor.execute("""
            SELECT COUNT(*)
            FROM information_schema.tables
            WHERE table_name = '{0}'
            """.format(tablename.replace('\'', '\'\'')))

        if cursor.fetchone()[0] == 1:
            return True

        return False


import csv
from os import path

def save_as_tsv(location, file_name, data):
    """
    Saves infomation in an ordered dictionary to a tsv file

    Parameters
    ----------
    location: str
        Absolute file path of the location to write tsv.
    file_name: str
        Name of the tsv.
    data: ordered dict
        Data to write to the tsv.
    """
    full_file_path = location + file_name + ".tsv"

    if path.exists(full_file_path):
        with open(full_file_path, "a+") as f:
            tsv_writer = csv.writer(f, delimiter='\t')
            tsv_writer.writerow(data.values())
    else:
        with open(full_file_path, "w", newline='') as f:
            tsv_writer = csv.writer(f, delimiter='\t')
            tsv_writer.writerow(data.keys())
            tsv_writer.writerow(data.values())